/** Piotr Stepaniak - E1I4

Sowniczek

		ODBC -	Interfejs pozwalajcy programom czy si z systemami zarzdzajcymi bazami danych. 
				Jest to API niezalene od jzyka programowania, systemu operacyjnego i bazy danych. (Wikipedia)
		
		SQLite - Baza danych, ktrej cech charakterystyczn jest brak rozbudowanej architektury klient-serwer. 
				Dane s przechowywane w plikach binarnych, a operacje SQL-owe s wykonywane przy uyciu doczonej 
				biblioteki dll lub sterownika ODBC. Bezpieczestwo przechowywanych danych zaley od bezpieczestwa 
				zapewnianego przez system plikw systemu operacyjnego. 
				(http://www.sqlite.org/)

Wymagania wstpne

		Aby skorzysta z poniszego przykadu naley:

		1) Zainstalowa sterowniki ODBC dla bazy danych SQLite (http://www.ch-werner.de/sqliteodbc/)
		2) Zaopatrzy si w wersj 4.x otl-a (http://otl.sourceforge.net/otl3_down.htm).
		3) Stworzy przykadowe poczenie ODBC o parametrach opisanych w staej DEFAULT_CONNECTION, 
		zgodnie z konwencj (login/password@user_dsn_name), gdzie user_dsn_name to nazwa poczenia 
		ODBC.

ODBC w Windows XP

		Aby stworzy wasne poczenie ODBC w systemie Windows XP, naley:

		a) Wybra 'Panel sterowania', a nastpnie 'Narzdzia administracjne'
		b) Wybra 'rda danych (ODBC)'
		c) W zakadce 'User DSN' naisn przycisk 'Add'
		d) Z wybranej listy wybra sterownik SQLite
		e) W nastpnym formularzu okreli nazw poczenia, a take plik zawierajcy baz danych.

Baza danych SQLite
		
		Aby stworzy baz danych SQLite mona posuy si konsolowym programem dostpnym na stronie projektu.

		Wystarczy wwczas wpisa:

			sqlite3 nazwa_bazy_danych

		aby stworzy plik zawierajcy baz danych o podanej nazwie. Po wpisaniu powyszej komendy uruchomi si
		konsola sql-owa. Komenda .help wywietla dostpne dodatkowe polecenia.

		Innym przydatnym narzdziem jest sqlite analyzer rwnie dostpny na stronie projektu. Udostpnia on 
		midzy innymi graficzn reprezentacj tabel, a take edytor zapyta sql.

Opis przykadu
		
		Poniszy przykad pokazuje jak przy uyciu sqlite'a i otl'a stworzy proste mapowanie:

			relacja (w sensie relacyjnych baz danych) - klasa ( w sensie programowania obiektowego)

		Powysza technika uatwia projektowanie oprogramowania (wzorzec entity - control) i bywa czsto stosowana
		w programowaniu obiektowym (patrz np. Hibernate).
		

*/

#include <iostream>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <string>
#include <vector>
#include <iterator>

#define OTL_ODBC // korzystaj z odbc 
#define OTL_STL  // udostpnienie funkcjonalnoci STL-owej 

//#define ODBCVER 0x0250 // Za pomoc tej komendy mona okreli wersj sterownika otl_odbc

#define DEFAULT_CONNECTION "scott/tiger@test3" // dane poczenia (user/password@odbc_name)

#include "otlv4.h"

using namespace std;


/** 
Klasa suca do przechowywania daty rozpoczcia pracy przez danego pracownika
*/
class Date {

	public:
		Date() {
			year = 1; month = 1; day = 1;
		};
		
		Date(const Date &d) {
			year= d.year; month = d.month; day = d.day;
		}

		Date(int y, int m, int d) : year(y), month(m), day(d) {};
	
		int getYear() const {
			return year;
		}

		int getMonth() const{
			return month;
		}

		int getDay() const {
			return day;
		}

	private:
		int year;
		int month;
		int day;
};

/**
Klasa reprezentujca pracownika
*/
class Employee {

	public:
		Employee (int i, const char* fname, const char* lname, int dId, int s, Date sDate) : 
			id(i), firstname(fname), lastname(lname), departmentId(dId), salary(s), startDate(sDate) {
		};

		Employee() {};

		Employee(const Employee &emp) {
			id = emp.getId();
			firstname = emp.getFirstname();
			lastname = emp.getLastname();
			departmentId = emp.getDepartmentId();
			salary = emp.getSalary();
			startDate = emp.getDate();
		}

		~Employee() {};

		void setId(int i) {
			id = i;
		}

		void setFirstname(string fname) {
			firstname = fname;
		}

		void setLastname(string lname) {
			lastname = lname;
		}

		void setDepartmentId(int dId) {
			departmentId = dId;
		}

		void setSalary(int s) {
			salary = s;
		}

		void setStartDate(Date sDate) {
			startDate = sDate;
		}

		int getId() const {
			return id;
		}

		const char* getFirstname() const {
			return firstname.c_str();
		}

		const char* getLastname() const {
			return lastname.c_str();
		}

		int getSalary() const {
			return salary;
		}

		int getDepartmentId() const {
			return departmentId;
		}

		Date getDate() const {
			return startDate;
		}

		Employee& operator=(const Employee &emp) {
			id = emp.getId();
			firstname = emp.getFirstname();
			lastname = emp.getLastname();
			departmentId = emp.getDepartmentId();
			salary = emp.getSalary();
			startDate = emp.getDate();
			return *this;
		}

		
		/**
		Przecienie operatora << dla strumienia typu otl_stream.
		Uywany przy pobieraniu kolekcji pracownikw z bazy (getEmployees).
		*/
		friend otl_stream& operator>>(otl_stream& s, Employee& emp) {
			
			char firstname[31],lastname[31];
			int id,departmentId, salary;
			otl_datetime date;
			bool create = false;
			
			s>>id;
			s>>firstname;
			s>>lastname;
			s>>departmentId;
			s>>salary;
			s>>date;
				
			emp.setId(id);
			emp.setFirstname(firstname);
			emp.setLastname(lastname);
			emp.setDepartmentId(departmentId);
			emp.setSalary(salary);
			emp.setStartDate(Date(date.year,date.month,date.month));

			return s;
		}

		/**
		Przecienie operatora << dla strumienia typu otl_stream.
		Uywany w operacjach insert.
		*/
		friend otl_stream& operator<< (otl_stream& s, const Employee& emp) {
			Date date = emp.getDate();
			otl_datetime datetime(date.getYear(),date.getMonth(),date.getDay(),0,0,0);  

			s<<emp.getId()
			 <<emp.getFirstname()
			 <<emp.getLastname()
			 <<emp.getDepartmentId()
			 <<emp.getSalary()
			 <<datetime;

			return s;
		}


		friend ostream& operator<< (ostream& out,const Employee &emp) {
			out << "Id: " << emp.getId() << ", " 
				<< "Imie: " << emp.getFirstname() << ", " 
				<< "Nazwisko: " <<emp.getLastname() << ", "
				<< "Id dzialu: " <<emp.getDepartmentId() << ", "
				<< "Pensja: " <<emp.getSalary()  <<", " 
				<< "Data zatrudnienia: " << emp.getDate().getYear() << "-" << emp.getDate().getMonth() << "-" <<emp.getDate().getDay() <<endl;

			return out;
		}

	private:
			int id;
			string firstname;
			string lastname;
			int	departmentId;
			int salary;
			Date startDate;
};


/**
Klasa zapewniajca poczenie z baz danych
*/
class EmployeeBusinessObject {

	public:

		EmployeeBusinessObject (const char* connString) {
			otl_connect::otl_initialize();  // inicjalizacja otl-a
			dbConn.rlogon(connString);      // ustanowienie poczenia zgodnie z podanym connString
			initDB();						// inicjalizacja bazy danych
		}

		~EmployeeBusinessObject() {
			dbConn.logoff();				// zamknicie poczenia z baz danych
		}

		// dodanie pracownika do relacji EmployeesTest
		void addEmployee(Employee *emp) {

			try {

				 otl_stream
					o(50, // wielko buffora po ktrej przekroczeniu nastpi jego oprnienie
					"insert into EmployeesTest values(:id<int>,:firstname<char[31]>,:lastname<char[31]>,:departmentId<int>,:salary<int>,:startDate<timestamp>)", 
					  dbConn  // obiekt poczenia
					 );

					//zapisanie danych pracownika do strumienia 
					o << *emp;
					// Alternatywna wersja kodu:	
					/*
					Date date = emp->getDate();
					otl_datetime datetime(date.getYear(),date.getMonth(),date.getDay(),0,0,0);  

					o<<emp->getId()
					 <<emp->getFirstname()
					 <<emp->getLastname()
					 <<emp->getDepartmentId()
					 <<emp->getSalary()
					 <<datetime;*/
					 
			} catch(otl_exception& p){ 
				cerr<<p.msg<<endl;		// wywietla wiadomo o bdzie
				cerr<<p.stm_text<<endl; // wywietla SQL, ktry spowodowa bd
				cerr<<p.sqlstate<<endl; // wywietla wiadomo SQLSTATE 
				cerr<<p.var_info<<endl; // wywietla zmienn, ktra spowodowaa bd
			} catch (std::exception &ex) { // obsuga wyjtkw standardowych 
				cerr << "An std:: exception occured" <<endl; 
				cerr << ex.what() <<endl;
			} catch (...) {                // obsuga pozostaych wyjtkwz
				cerr << "Unknown error occured" <<endl;
			}

		}

		// dodanie pracownikw zapisanych w kolekcji do relacji EmployeesTest
		void addEmployees(vector<Employee> &emps) {
			try {

				 otl_stream
					o(50, // rozmiar bufora
					"insert into EmployeesTest values(:id<int>,:firstname<char[31]>,:lastname<char[31]>,:departmentId<int>,:salary<int>,:startDate<timestamp>)", 
					  dbConn  // obiekt poczenia
					 );

				 // przekopiowanie zawartoci kolekcji do strumienia o
				 // wykorzystywane s przy tym przecione operatory << klasy Employee
				 copy(emps.begin(), 
					 emps.end(), 
					 otl_output_iterator<Employee>(o)
				 );	 
					 
			} catch(otl_exception& p){ 
				cerr<<p.msg<<endl;		// wywietla wiadomo o bdzie
				cerr<<p.stm_text<<endl; // wywietla SQL, ktry spowodowa bd
				cerr<<p.sqlstate<<endl; // wywietla wiadomo SQLSTATE 
				cerr<<p.var_info<<endl; // wywietla zmienn, ktra spowodowaa bd
			} catch (std::exception &ex) { // obsuga wyjtkw standardowych 
				cerr << "An std:: exception occured" <<endl; 
				cerr << ex.what() <<endl;
			} catch (...) {                // obsuga pozostaych wyjtkwz
				cerr << "Unknown error occured" <<endl;
			}

		}

		// uaktualnienie danych o podanym pracowniku
		void updateEmployee(Employee *emp) {

			try {

				otl_stream 
				   o(1, // rozmiar bufora
					 "UPDATE EmployeesTest "
					 "   SET firstname=:firstname<char[31]>,lastname=:lastname<char[31]>,departmentId=:departmentId<int>,salary=:salary<int>,startDate=:startDate<timestamp> "
					 " WHERE id=:id<int>", 
					 dbConn // obiekt poczenia
					);

					// odwzorowanie obiektu Date na otl_datetime
					Date date = emp->getDate();
					otl_datetime datetime(date.getYear(),date.getMonth(),date.getDay(),0,0,0);  

					// zapisanie danych do strumienia
				   o<<emp->getFirstname()
					<<emp->getLastname()
					<<emp->getDepartmentId()
					<<emp->getSalary()
					<<datetime
					<<emp->getId();
					

			} catch(otl_exception& p){ 
				cerr<<p.msg<<endl;		// wywietla wiadomo o bdzie
				cerr<<p.stm_text<<endl; // wywietla SQL, ktry spowodowa bd
				cerr<<p.sqlstate<<endl; // wywietla wiadomo SQLSTATE 
				cerr<<p.var_info<<endl; // wywietla zmienn, ktra spowodowaa bd
			} catch (std::exception &ex) { // obsuga wyjtkw standardowych 
				cerr << "An std:: exception occured" <<endl; 
				cerr << ex.what() <<endl;
			} catch (...) {                // obsuga pozostaych wyjtkwz
				cerr << "Unknown error occured" <<endl;
			}

		}

		// usunicie pracownika o podanym id z relacji EmployeesTest
		void removeEmployee(int id) {

			try {

				otl_stream 
				   o(1, // rozmiar bufora
					 " DELETE FROM EmployeesTest "
					 " WHERE id=:id<int>", 
					 dbConn // obiekt poczenia
					);

				// przekazanie parametru do sql-a
				o<<id;

			} catch(otl_exception& p){ 
				cerr<<p.msg<<endl;		// wywietla wiadomo o bdzie
				cerr<<p.stm_text<<endl; // wywietla SQL, ktry spowodowa bd
				cerr<<p.sqlstate<<endl; // wywietla wiadomo SQLSTATE 
				cerr<<p.var_info<<endl; // wywietla zmienn, ktra spowodowaa bd
			} catch (std::exception &ex) { // obsuga wyjtkw standardowych 
				cerr << "An std:: exception occured" <<endl; 
				cerr << ex.what() <<endl;
			} catch (...) {                // obsuga pozostaych wyjtkwz
				cerr << "Unknown error occured" <<endl;
			}

		}

		// pobranie informacji nt. pracownika o podanym id z relacji EmployeesTest 
		Employee* getEmployee(const int id) {

			Employee *emp = NULL;

			try {

				 otl_stream i(50, // rozmiar bufora
							  "select firstname,lastname,departmentId,salary,startDate from EmployeesTest "
							  "where id=:id<int> ",
							  dbConn // obiekt poczenia
							 ); 
				 
				 char firstname[31],lastname[31];
				 int departmentId, salary;
				 otl_datetime date;
				 bool create = false;

				 // przekazanie parametru do sql-a
				 i<<id; 

				 // dopki istniej dane w strumieniu - powtarzaj
				 // poniewa id pracownikw s unikalne, ptla bdzie miaa zawsze 
				 // albo jeden albo zero obiegw
				 while(!i.eof()){ 
				  create = true;
				  i>>firstname;
				  i>>lastname;
				  i>>departmentId;
				  i>>salary;
				  i>>date;
				 }
				
				 // jeli istnia pracownik o podanym id
				 if (create)
					 emp = new Employee(id,
										firstname,
										lastname,
										departmentId,
										salary,
										Date(date.year,date.month,date.month));

			} catch(otl_exception& p){ 
				cerr<<p.msg<<endl;		// wywietla wiadomo o bdzie
				cerr<<p.stm_text<<endl; // wywietla SQL, ktry spowodowa bd
				cerr<<p.sqlstate<<endl; // wywietla wiadomo SQLSTATE 
				cerr<<p.var_info<<endl; // wywietla zmienn, ktra spowodowaa bd
			} catch (std::exception &ex) { // obsuga wyjtkw standardowych 
				cerr << "An std:: exception occured" <<endl; 
				cerr << ex.what() <<endl;
			} catch (...) {                // obsuga pozostaych wyjtkwz
				cerr << "Unknown error occured" <<endl;
			}

			return emp;
		}

		// pobranie informacji nt. pracownikw z relacji EmployeesTest
		vector<Employee> getEmployees() {
			vector<Employee> v; // vector of rows
			
			try {		
				otl_stream i(50, //rozmiar bufora
							  "select * from EmployeesTest ",
							  dbConn // obiekt poczenia
							 ); 
				 

				// przekopiuj do wektora wszystkie dane wyjciowe zawarte w strumieniu
				// wykorzystywany jest tu przeciony operator >> klasy Employee
				copy(otl_input_iterator<Employee,ptrdiff_t>(i), 
					  otl_input_iterator<Employee,ptrdiff_t>(),
					  back_inserter(v));    

			} catch(otl_exception& p){ 
				cerr<<p.msg<<endl;		// wywietla wiadomo o bdzie
				cerr<<p.stm_text<<endl; // wywietla SQL, ktry spowodowa bd
				cerr<<p.sqlstate<<endl; // wywietla wiadomo SQLSTATE 
				cerr<<p.var_info<<endl; // wywietla zmienn, ktra spowodowaa bd
			} catch (std::exception &ex) { // obsuga wyjtkw standardowych 
				cerr << "An std:: exception occured" <<endl; 
				cerr << ex.what() <<endl;
			} catch (...) {                // obsuga pozostaych wyjtkwz
				cerr << "Unknown error occured" <<endl;
			}
			
			return v;
		}

		// sprawdza czy istnieje pracownik o zadanym id
		bool existEmployee(int id) {

			try {

				 otl_stream i(50, // rozmiar bufora
							  "select count(*) from EmployeesTest "
							  "where id=:id<int> ",
							  dbConn // obiekt poczenia
							 ); 

				 char count = 0;

				 // zapidanie parametrw do sql-a
				 i<<id; 
			
				 // dopki s dane - powtarzaj
				 // ptla ta bdzie miaa tylko jeden obieg
				 while(!i.eof()){ 
				  i>>count;
				 }
				
				 // jeli liczba pracownik o zadnym id jest rna od 0
				 if (count != '0') return true;

			} catch(otl_exception& p){ 
				cerr<<p.msg<<endl;		// wywietla wiadomo o bdzie
				cerr<<p.stm_text<<endl; // wywietla SQL, ktry spowodowa bd
				cerr<<p.sqlstate<<endl; // wywietla wiadomo SQLSTATE 
				cerr<<p.var_info<<endl; // wywietla zmienn, ktra spowodowaa bd
			} catch (std::exception &ex) { // obsuga wyjtkw standardowych 
				cerr << "An std:: exception occured" <<endl; 
				cerr << ex.what() <<endl;
			} catch (...) {                // obsuga pozostaych wyjtkwz
				cerr << "Unknown error occured" <<endl;
			}

			return false;
		}


	private:
		/**
		Inicjalizacja bazy danych.
		*/
		void initDB() {

			try {

			  // usunicie relacji EmployeesTest
			  otl_cursor::direct_exec
			   (
				dbConn,
				"drop table EmployeesTest",
				otl_exception::disabled // wyczenie wyjtkw OTL
			   ); 

			  // stworzenie relacji EmployeesTest
			  otl_cursor::direct_exec
			   (
				dbConn,
				"create table EmployeesTest(id integer primary key autoincrement,"
				"firstname varchar(30),"
				"lastname varchar(30),"
				"departmentId int,"
				"salary int,"
				"startDate date)"
				); 

			} catch(otl_exception& p){ 
				cerr<<p.msg<<endl;		// wywietla wiadomo o bdzie
				cerr<<p.stm_text<<endl; // wywietla SQL, ktry spowodowa bd
				cerr<<p.sqlstate<<endl; // wywietla wiadomo SQLSTATE 
				cerr<<p.var_info<<endl; // wywietla zmienn, ktra spowodowaa bd
			} catch (std::exception &ex) { // obsuga wyjtkw standardowych 
				cerr << "An std:: exception occured" <<endl; 
				cerr << ex.what() <<endl;
			} catch (...) {                // obsuga pozostaych wyjtkwz
				cerr << "Unknown error occured" <<endl;
			}

		}

	private:
		otl_connect dbConn; // obiekt poczenia

};

// interfejs obiektu dostarczajcego informacji o pracownikach
virtual class Employees {

	public:

		virtual void add(Employee *emp) = 0;

		virtual void add(vector<Employee> &emps) = 0;

		virtual void remove(int id) = 0;

		virtual void update(Employee *emp) = 0;

		virtual Employee* get(int id) = 0;

		virtual vector<Employee> getAll() = 0;

		virtual bool exist(int id) = 0;

		friend ostream& operator<< (ostream &out, Employees &emps); 

};

// implementacja interfejsu Employees - z wykorzystaniem bazy danych
class EmployeesDB : public Employees {

	public:

		EmployeesDB() {
			// stwrz pomocniczy obiekt biznesowy
			empBO = new EmployeeBusinessObject(DEFAULT_CONNECTION);
		}

		// dodaj pracownika
		void add(Employee *emp) {
			empBO->addEmployee(emp);
		}

		// dodaj pracownikw
		void add(vector<Employee> &emps) {
			empBO->addEmployees(emps);
		}

		// usu pracownika o podanym id
		void remove(int id) {
			empBO->removeEmployee(id);
		}

		// uaktualnij podanego pracownika
		void update(Employee *emp) {
			empBO->updateEmployee(emp);
		}

		// zwr informacje o pracowniku o podanym id
		Employee* get(int id) {
			return empBO->getEmployee(id);
		}

		// zwr kolekcj zawierajc informacje o wszystkich pracownikach
		vector<Employee> getAll() {
			return empBO->getEmployees();
		}

		// sprawd czy istnieje pracownik o podanym id
		bool exist (int id) {
			return empBO->existEmployee(id);
		}

		// przecienie strumienia << dla potrzeb np. serializacji
		friend ostream& operator<< (ostream &out, Employees &emps) {
			// pobierz kolekcje wszystkich pracownikw
			vector<Employee> empsVec = emps.getAll();
			// przekopiuj pobran kolekcj do strumienia out
			copy(empsVec.begin(), empsVec.end(), ostream_iterator<Employee>(out, "\n"));

			return out;
		}

	private:
	
		EmployeeBusinessObject *empBO;

};



int main () {

	cout << "Otl sqlite example" << endl << endl;
	
	cout << "Adding an employee" << endl;
	Employees *emps = new EmployeesDB(); // stworzenie obiektu EmployeesDB polczonego z relacj EmployeesTest
	Date date(2006,6,15);
	Employee *emp1 = new Employee(1, "Piotr", "Stepaniak", 1, 1250, date); // tworzenie nowego pracownika
	emps->add( emp1 ); // dodanie pracownika do bazy
	cout << *emps->get(1) <<endl;  // pobranie inf. o pracowniku z bazy

	cout << "Changing employees salary " << endl;
	emp1->setSalary(1500);
	emps->update(emp1); // zmiana informacji o danym pracowniku
	cout << *emps->get(1) <<endl;

	cout << "Adding 2 more employees" << endl;
	vector<Employee> vec;
	vec.push_back(*(new Employee(2, "Pawe", "Kozak", 1, 1250, date)));
	vec.push_back(*(new Employee(3, "Tomasz", "Ziss", 1, 1250, date)));
	emps->add(vec); // dodanie kolekcji nowych pracownikw do bazy
	
	cout<< "Employees nmbr: " << (emps->getAll()).size()<<endl; // pobranie kolekcji z wszystkimi pracownikami zapisanymi w bazie
																// i wywietlenie jej rozmiaru
	cout<< "Employees list: " << endl << endl;
	cout<<*emps; // wywietlenie wszystkich pracownikw zapisanych w bazie

	cout << "Checking if employee 2 is present" <<endl;
	if (emps->exist(2)) { // sprawdzenie czy istnieje w bazie pracownik o id = 2
		cout << "There is such an employee!" <<endl;
	} else 
		cout << "There is no such employee!" <<endl;

	cout << "Deleting employee 2" << endl;
	emps->remove(2); // usunicie pracownika o id = 2

	cout << "Checking if employee 2 is present" <<endl;
	if (emps->exist(2)) { // sprawdzenie czy istnieje w bazie pracownik o id = 2
		cout << "There is such an employee!" <<endl;
	} else 
		cout << "There is no such employee!" <<endl;

	cout<< endl <<"Employees nmbr: " << (emps->getAll()).size()<<endl;
	cout<< "Employees list: " << endl << endl;
	cout<<*emps;

	cout << "Press a key .. " <<endl;
	char c;
	cin >> c;
	return 0;
}
